
import os
import sys
import logging
import logging.handlers
import enum
from argparse import ArgumentParser, RawTextHelpFormatter, FileType, SUPPRESS
from .gitlab_tree import GitlabTree
from .format import PrintFormat
from .method import CloneMethod
from .naming import FolderNaming
from .archive import ArchivedResults
from . import __version__ as VERSION

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
log = logging.getLogger(__name__)

def main():
    args = parse_args(argv=None if sys.argv[1:] else ['--help'])
    if args.version:
        print(VERSION)
        sys.exit(0) 

    if args.token is None:
        print('Please specify a valid token with the -t flag or the \'GITLAB_TOKEN\' environment variable')
        sys.exit(1)

    if args.url is None:
        print('Please specify a valid gitlab base url with the -h flag or the \'GITLAB_URL\' environment variable')
        sys.exit(1)

    elif args.dest is None and args.print is False:
        print('Please specify a destination for the gitlab tree')
        sys.exit(1)

    config_logging(args)
    includes=split(args.include)
    excludes=split(args.exclude)

    tree = GitlabTree(args.url, args.token, args.method, args.naming, args.archived.api_value, includes,
                      excludes, args.file, args.concurrency, args.recursive, args.verbose)
    log.debug("Reading projects tree from gitlab at [%s]", args.url)
    tree.load_tree()

    if tree.is_empty():
        log.fatal("The tree is empty, check your include/exclude patterns or run with more verbosity for debugging")
        sys.exit(1) 

    if args.print:
        tree.print_tree(args.print_format)
    else:
        tree.sync_tree(args.dest)


def split(arg):
    return arg.split(",") if arg != "" else None

def config_logging(args):
    if args.verbose:
        handler = logging.StreamHandler(sys.stdout)
        logging.root.handlers = []
        handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
        logging.root.addHandler(handler)
        level = logging.ERROR if args.print else logging.DEBUG
        logging.root.setLevel(level)
        log.debug("verbose=[%s], print=[%s], log level set to [%s] level", args.verbose, args.print, level)
        os.environ["GIT_PYTHON_TRACE"] = 'full'


def parse_args(argv=None):
    example_text = '''examples:

    clone an entire gitlab tree using a url and a token:
    gitlabber -t <personal access token> -u <gitlab url>

    only print the gitlab tree:
    gitlabber -p .

    clone only projects under subgroup 'MySubGroup' to location '~/GitlabRoot':
    gitlabber -i '/MyGroup/MySubGroup**' ~/GitlabRoot

    clone only projects under group 'MyGroup' excluding any projects under subgroup 'MySubGroup':
    gitlabber -i '/MyGroup**' -x '/MyGroup/MySubGroup**' .

    clone an entire gitlab tree except projects under groups named 'ArchiveGroup':
    gitlabber -x '/ArchiveGroup**' .

    clone projects that start with a case insensitive 'w' using a regular expression:
    gitlabber -i '/{[w].*}' .
    '''

    parser = ArgumentParser(
        description='Gitlabber - clones or pulls entire groups/projects tree from gitlab',
        prog="gitlabber",
        epilog=example_text,
        formatter_class=RawTextHelpFormatter)    
    parser.add_argument(
        'dest',
        nargs='?', 
        type=validate_path,
        help='destination path for the cloned tree (created if doesn\'t exist)')
    parser.add_argument(
        '-t',
        '--token',
        metavar=('token'),
        default=os.environ.get('GITLAB_TOKEN'),
        help='gitlab personal access token https://docs.gitlab.com/ee/user/profile/personal_access_tokens.html')
    parser.add_argument(
        '-u',
        '--url',
        metavar=('url'),
        default=os.environ.get('GITLAB_URL'),
        help='base gitlab url (e.g.: \'http://gitlab.mycompany.com\')')
    parser.add_argument(
        '--verbose',
        action='store_true',
        help='print more verbose output')
    parser.add_argument(
        '-f',
        '--file',
        metavar=('file'),
        help=SUPPRESS)    
    parser.add_argument(
        '-c',
        '--concurrency',
        default=os.environ.get('GITLABBER_GIT_CONCURRENCY', 1),
        type=int,
        metavar=('concurrency'),
        help=SUPPRESS)
    parser.add_argument(
        '-p',
        '--print',
        action='store_true',
        help='print the tree without cloning')
    parser.add_argument(
        '--print-format', 
        type=PrintFormat.argparse,
        default=PrintFormat.TREE,
        choices=list(PrintFormat),
        help='print format (default: \'tree\')')
    parser.add_argument(
        '-n',
        '--naming',
        type=FolderNaming.argparse,
        choices=list(FolderNaming),
        default=FolderNaming.argparse(os.environ.get('GITLABBER_FOLDER_NAMING', "name")),
        help='the folder naming strategy for projects from the gitlab API attributes (default: "name")')
    parser.add_argument(
        '-m',
        '--method',
        type=CloneMethod.argparse,
        choices=list(CloneMethod),
        default=os.environ.get('GITLABBER_CLONE_METHOD', "ssh"),
        help='the git transport method to use for cloning (default: "ssh")')
    parser.add_argument(
        '-a',
        '--archived',
        type=ArchivedResults.argparse,
        choices=list(ArchivedResults),
        default=ArchivedResults.INCLUDE,
        help='include archived projects and groups in the results (default: "include")')
    parser.add_argument(
        '-i',
        '--include',
        metavar=('csv'),
        default=os.environ.get('GITLABBER_INCLUDE', ""),
        help='comma delimited list of glob patterns of paths to projects or groups to clone/pull')
    parser.add_argument(
        '-x',
        '--exclude',
        metavar=('csv'),
        default=os.environ.get('GITLABBER_EXCLUDE', ""),
        help='comma delimited list of glob patterns of paths to projects or groups to exclude from clone/pull')
    parser.add_argument(
        '-r',
        '--recursive',
        action='store_true',
        default=False,
        help='clone/pull git submodules recursively')
    parser.add_argument(
        '--version',
        action='store_true',
        help='print the version')

    args = parser.parse_args(argv)
    args_print = vars(args).copy()
    args_print['token'] = 'xxxxx'
    log.debug("running with args [%s]", args_print)
    return args

def validate_path(value):
    if value.endswith('/'):
        return value[:-1]
    return value




from gitlabber.archive import ArchivedResults
import pytest
import re

def test_archive_parse():
    assert ArchivedResults.INCLUDE == ArchivedResults.argparse("include")

def test_archive_string():
    assert "exclude" == ArchivedResults.__str__(ArchivedResults.EXCLUDE)

def test_repr():
    retval = repr(ArchivedResults.ONLY)
    match = re.match("^<ArchivedResults: ({.*})>$", retval)

def test_archive_api_value():
    assert True == ArchivedResults.ONLY.api_value
    assert False == ArchivedResults.EXCLUDE.api_value
    assert None == ArchivedResults.INCLUDE.api_value

def test_archive_invalid():
    assert "invalid_value" == ArchivedResults.argparse("invalid_value")
        



import enum

class ArchivedResults(enum.Enum):
    INCLUDE = (1, None)
    EXCLUDE = (2, False)
    ONLY = (3, True)

    def __init__(self, int_value, api_value):
        self.int_value = int_value
        self.api_value = api_value 

    def __str__(self):
        return self.name.lower()

    def __repr__(self):
        return str(self)

    @staticmethod
    def argparse(s):
        try:
            return ArchivedResults[s.upper()]
        except KeyError:
            return s



import enum 

class PrintFormat(enum.IntEnum):
    JSON = 1
    YAML = 2
    TREE = 3

    def __str__(self):
        return self.name.lower()

    def __repr__(self):
        return str(self)

    @staticmethod
    def argparse(s):
        try:
            return PrintFormat[s.upper()]
        except KeyError:
            return s



from gitlab import Gitlab
from anytree import Node, RenderTree
from anytree.exporter import DictExporter, JsonExporter
from anytree.importer import DictImporter
from .git import sync_tree
from .format import PrintFormat
from .method import CloneMethod
from .naming import FolderNaming
from .progress import ProgressBar
import yaml
import globre
import logging
import os
log = logging.getLogger(__name__)


class GitlabTree:

    def __init__(self, url, token, method, naming, archived=None, includes=[], excludes=[], in_file=None, concurrency=1, recursive=False, disable_progress=False):
        self.includes = includes
        self.excludes = excludes
        self.url = url
        self.root = Node("", root_path="", url=url)
        self.gitlab = Gitlab(url, private_token=token,
                             ssl_verify=GitlabTree.get_ca_path())
        self.method = method
        self.naming = naming
        self.archived = archived
        self.in_file = in_file
        self.concurrency = concurrency
        self.recursive = recursive
        self.disable_progress = disable_progress
        self.progress = ProgressBar('* loading tree', disable_progress)

    @staticmethod
    def get_ca_path():
        """
        returns REQUESTS_CA_BUNDLE, CURL_CA_BUNDLE, or True
        """
        return next(item for item in [os.getenv('REQUESTS_CA_BUNDLE', None), os.getenv('CURL_CA_BUNDLE', None), True]
                    if item is not None)

    def is_included(self, node):
        '''
        returns True if the node should be included.
        if there are no include patterns then everything is included
        any include patterns matching the root path will result in inclusion
        '''
        if self.includes is not None:
            for include in self.includes:
                if globre.match(include, node.root_path):
                    log.debug(
                        "Matched include path [%s] to node [%s]", include, node.root_path)
                    return True
        else:
            return True

    def is_excluded(self, node):
        '''
        returns True if the node should be excluded 
        if the are no exclude patterns then nothing is excluded
        any exclude pattern matching the root path will result in exclusion
        '''
        if self.excludes is not None:
            for exclude in self.excludes:
                if globre.match(exclude, node.root_path):
                    log.debug(
                        "Matched exclude path [%s] to node [%s]", exclude, node.root_path)
                    return True
        return False

    def filter_tree(self, parent):
        for child in parent.children:
            if not child.is_leaf:
                self.filter_tree(child)
                if child.is_leaf:
                    if not self.is_included(child):
                        child.parent = None
                    if self.is_excluded(child):
                        child.parent = None
            else:
                if not self.is_included(child):
                    child.parent = None
                if self.is_excluded(child):
                    child.parent = None

    def root_path(self, node):
        return "/".join([str(n.name) for n in node.path])

    def make_node(self, name, parent, url):
        node = Node(name=name, parent=parent, url=url)
        node.root_path = self.root_path(node)
        return node

    def add_projects(self, parent, projects):
        for project in projects:
            project_id = project.name if self.naming == FolderNaming.NAME else project.path
            project_url = project.ssh_url_to_repo if self.method is CloneMethod.SSH else project.http_url_to_repo
            node = self.make_node(project_id, parent,
                                  url=project_url)
            self.progress.show_progress(node.name, 'project')

    def get_projects(self, group, parent):
        projects = group.projects.list(as_list=False, archived=self.archived)
        self.progress.update_progress_length(len(projects))
        self.add_projects(parent, projects)

    def get_subgroups(self, group, parent):
        subgroups = group.subgroups.list(as_list=False, archived=self.archived)
        self.progress.update_progress_length(len(subgroups))
        for subgroup_def in subgroups:
            subgroup = self.gitlab.groups.get(subgroup_def.id)
            subgroup_id = subgroup.name if self.naming == FolderNaming.NAME else subgroup.path
            node = self.make_node(subgroup_id, parent, url=subgroup.web_url)
            self.progress.show_progress(node.name, 'group')
            self.get_subgroups(subgroup, node)
            self.get_projects(subgroup, node)

    def load_gitlab_tree(self):
        groups = self.gitlab.groups.list(as_list=False, archived=self.archived)
        self.progress.init_progress(len(groups))
        for group in groups:
            if group.parent_id is None:
                group_id = group.name if self.naming == FolderNaming.NAME else group.path
                node = self.make_node(group_id, self.root, url=group.web_url)
                self.progress.show_progress(node.name, 'group')
                self.get_subgroups(group, node)
                self.get_projects(group, node)

        elapsed = self.progress.finish_progress()
        log.debug("Loading projects tree from gitlab took [%s]", elapsed)

    def load_file_tree(self):
        with open(self.in_file, 'r') as stream:
            dct = yaml.safe_load(stream)
            self.root = DictImporter().import_(dct)

    def load_tree(self):
        if self.in_file:
            log.debug("Loading tree from file [%s]", self.in_file)
            self.load_file_tree()
        else:
            log.debug("Loading projects tree gitlab server [%s]", self.url)
            self.load_gitlab_tree()

        log.debug("Fetched root node with [%d] projects" % len(
            self.root.leaves))
        self.filter_tree(self.root)

    def print_tree(self, format=PrintFormat.TREE):
        if format is PrintFormat.TREE:
            self.print_tree_native()
        elif format is PrintFormat.YAML:
            self.print_tree_yaml()
        elif format is PrintFormat.JSON:
            self.print_tree_json()
        else:
            log.fatal("Invalid print format [%s]", format)

    def print_tree_native(self):
        for pre, _, node in RenderTree(self.root):
            line = ""
            if node.is_root:
                line = "%s%s [%s]" % (pre, "root", self.url)
            else:
                line = "%s%s [%s]" % (pre, node.name, node.root_path)
            print(line)

    def print_tree_yaml(self):
        dct = DictExporter().export(self.root)
        print(yaml.dump(dct, default_flow_style=False))

    def print_tree_json(self):
        exporter = JsonExporter(indent=2, sort_keys=True)
        print(exporter.export(self.root))

    def sync_tree(self, dest):
        log.debug("Going to clone/pull [%s] groups and [%s] projects" %
                  (len(self.root.descendants) - len(self.root.leaves), len(self.root.leaves)))
        sync_tree(self.root, dest, concurrency=self.concurrency,
                  disable_progress=self.disable_progress, recursive=self.recursive)

    def is_empty(self):
        return self.root.height < 1




import tests.gitlab_test_utils as gitlab_util
import tests.io_test_util as output_util
from gitlabber.archive import ArchivedResults

def test_load_tree(monkeypatch):
    gl = gitlab_util.create_test_gitlab(monkeypatch)
    gl.load_tree()
    gl.print_tree()
    gitlab_util.validate_tree(gl.root)


def test_filter_tree_include_positive(monkeypatch):
    gl = gitlab_util.create_test_gitlab(monkeypatch, includes=["/group**"])
    gl.load_tree()
    gitlab_util.validate_tree(gl.root)


def test_filter_tree_include_negative(monkeypatch):
    gl = gitlab_util.create_test_gitlab(monkeypatch, includes=["/no_match**"])
    gl.load_tree()
    assert gl.root.is_leaf is True
    assert len(gl.root.children) == 0


def test_filter_tree_include_deep_positive(monkeypatch):
    gl = gitlab_util.create_test_gitlab(monkeypatch, includes=["/group/subgroup/project"])
    gl.load_tree()
    assert gl.root.is_leaf is False
    assert len(gl.root.children) == 1
    assert len(gl.root.children[0].children) == 1
    assert len(gl.root.children[0].children[0].children) == 1
    assert gl.root.children[0].children[0].children[0].is_leaf is True


def test_filter_tree_exclude_positive(monkeypatch):
    gl = gitlab_util.create_test_gitlab(monkeypatch, excludes=["/group**"])
    gl.load_tree()
    assert gl.root.is_leaf is True
    assert len(gl.root.children) == 0


def test_filter_tree_exclude_deep_positive(monkeypatch):
    gl = gitlab_util.create_test_gitlab(monkeypatch, excludes=["/group/subgroup/project**"])
    gl.load_tree()
    assert gl.root.is_leaf is False
    assert gl.root.height == 2
    assert gl.root.children[0].height == 1


def test_filter_tree_exclude_negative(monkeypatch):
    gl = gitlab_util.create_test_gitlab(monkeypatch, excludes=["/no_match**"])
    gl.load_tree()
    gitlab_util.validate_tree(gl.root)
    
def test_print_tree_json(monkeypatch):
    gl = gitlab_util.create_test_gitlab(monkeypatch)
    gl.load_tree()
    from gitlabber.format import PrintFormat
    import json
    with output_util.captured_output() as (out, err):
        gl.print_tree(PrintFormat.JSON)
        output = json.loads(out.getvalue())
        with open(gitlab_util.JSON_TEST_OUTPUT_FILE, 'r') as jsonFile:
            output_file = json.load(jsonFile)
            assert json.dumps(output_file, sort_keys=True, indent=2) == json.dumps(
                output, sort_keys=True, indent=2)


def test_print_tree_yaml(monkeypatch):
    gl = gitlab_util.create_test_gitlab(monkeypatch)
    gl.load_tree()
    from gitlabber.format import PrintFormat
    import yaml
    with output_util.captured_output() as (out, err):
        gl.print_tree(PrintFormat.YAML)
        output = yaml.safe_load(out.getvalue())
        with open(gitlab_util.YAML_TEST_OUTPUT_FILE, 'r') as yamlFile:
            output_file = yaml.safe_load(yamlFile)
            assert yaml.dump(output_file) == yaml.dump(output)


def test_load_tree_from_file(monkeypatch):
    gl = gitlab_util.create_test_gitlab(
        monkeypatch, in_file=gitlab_util.JSON_TEST_OUTPUT_FILE)
    gl.load_tree()
    gitlab_util.validate_tree(gl.root)


def test_empty_tree(monkeypatch):
    gl = gitlab_util.create_test_gitlab(monkeypatch, excludes=["/group**"])
    gl.load_tree()
    assert gl.is_empty() is True

def test_archive_included(monkeypatch):
    gl = gitlab_util.create_test_gitlab_with_archived(monkeypatch, archived=ArchivedResults.INCLUDE.api_value)
    gl.load_tree()
    assert gl.root.is_leaf is False
    assert len(gl.root.children) == 2
    assert len(gl.root.children[0].children) == 2
    assert len(gl.root.children[0].children[0].children) == 2

def test_archive_excluded(monkeypatch):
    gl = gitlab_util.create_test_gitlab_with_archived(monkeypatch, archived=ArchivedResults.EXCLUDE.api_value)
    
    gl.load_tree()
    assert gl.root.is_leaf is False
    assert len(gl.root.children) == 1
    assert len(gl.root.children[0].children) == 1
    assert len(gl.root.children[0].children[0].children) == 1
    
    assert "_archived_" not in gl.root.children[0].name
    assert "_archived_" not in gl.root.children[0].children[0].name
    assert "_archived_" not in gl.root.children[0].children[0].children[0].name

def test_archive_only(monkeypatch):
    gl = gitlab_util.create_test_gitlab_with_archived(monkeypatch, archived=ArchivedResults.ONLY.api_value)
    
    gl.load_tree()
    gl.print_tree()
    assert gl.root.is_leaf is False
    assert len(gl.root.children) == 1
    assert len(gl.root.children[0].children) == 1
    assert len(gl.root.children[0].children[0].children) == 1
    
    assert "_archived_" in gl.root.children[0].name
    assert "_archived_" in gl.root.children[0].children[0].children[0].name

def test_get_ca_path(monkeypatch):
    import os
    from gitlabber import gitlab_tree
    os.environ["REQUESTS_CA_BUNDLE"] = "/tmp"
    result = gitlab_tree.GitlabTree.get_ca_path()
    assert result == "/tmp"
    del os.environ['REQUESTS_CA_BUNDLE']

    os.environ["CURL_CA_BUNDLE"] = "/tmp2"
    result = gitlab_tree.GitlabTree.get_ca_path()
    assert result == "/tmp2"
    del os.environ['CURL_CA_BUNDLE']

    result = gitlab_tree.GitlabTree.get_ca_path()
    assert result == True






import os
import json
from gitlabber import __version__ as VERSION
import tests.gitlab_test_utils as gitlab_util
import tests.io_test_util as io_util
import pytest
import coverage
coverage.process_startup()


@pytest.mark.slow_integration_test
def test_clone_subgroup():
    os.environ['GITLAB_URL'] = 'http://www.gitlab.com/'
    output = io_util.execute(['-p', '--print-format', 'json'], 60)
    obj = json.loads(output)
    assert obj['children'][0]['name'] == 'Group Test'
    assert obj['children'][0]['children'][0]['name'] == 'Subgroup Test'
    assert len(obj['children'][0]['children'][0]['children']) == 3
    assert obj['children'][0]['children'][0]['children'][0]['name'] == 'archived-project'
    assert obj['children'][0]['children'][0]['children'][1]['name'] == 'gitlab-project-submodule'
    assert obj['children'][0]['children'][0]['children'][2]['name'] == 'gitlabber-sample-submodule'

@pytest.mark.slow_integration_test
def test_clone_subgroup_exclude_archived():
    os.environ['GITLAB_URL'] = 'http://www.gitlab.com/'
    output = io_util.execute(['-p', '--print-format', 'json', '-a', 'exclude'], 60)
    obj = json.loads(output)
    assert obj['children'][0]['name'] == 'Group Test'
    assert obj['children'][0]['children'][0]['name'] == 'Subgroup Test'
    assert len(obj['children'][0]['children'][0]['children']) == 2
    assert obj['children'][0]['children'][0]['children'][0]['name'] == 'gitlab-project-submodule'
    assert obj['children'][0]['children'][0]['children'][1]['name'] == 'gitlabber-sample-submodule'

@pytest.mark.slow_integration_test
def test_clone_subgroup_only_archived():
    os.environ['GITLAB_URL'] = 'http://www.gitlab.com/'
    output = io_util.execute(['-p', '--print-format', 'json', '-a', 'only'], 60)
    obj = json.loads(output)
    assert obj['children'][0]['name'] == 'Group Test'
    assert obj['children'][0]['children'][0]['name'] == 'Subgroup Test'
    assert len(obj['children'][0]['children'][0]['children']) == 1
    assert obj['children'][0]['children'][0]['children'][0]['name'] == 'archived-project'


@pytest.mark.slow_integration_test
def test_clone_subgroup_only_archived():
    os.environ['GITLAB_URL'] = 'http://www.gitlab.com/'
    output = io_util.execute(['-p', '--print-format', 'json', '-a', 'only'], 60)
    obj = json.loads(output)
    assert obj['children'][0]['name'] == 'Group Test'
    assert obj['children'][0]['children'][0]['name'] == 'Subgroup Test'
    assert len(obj['children'][0]['children'][0]['children']) == 1
    assert obj['children'][0]['children'][0]['children'][0]['name'] == 'archived-project'


@pytest.mark.slow_integration_test
def test_clone_subgroup_naming_path():
    os.environ['GITLAB_URL'] = 'http://www.gitlab.com/'
    output = io_util.execute(['-p', '--print-format', 'json', '-n', 'path'], 60)
    obj = json.loads(output)
    assert obj['children'][0]['name'] == 'erez-group-test'
    assert obj['children'][0]['children'][0]['name'] == 'subgroup-test'
    assert len(obj['children'][0]['children'][0]['children']) == 3
    assert obj['children'][0]['children'][0]['children'][0]['name'] == 'archived-project'
    assert obj['children'][0]['children'][0]['children'][1]['name'] == 'gitlab-project-submodule'
    assert obj['children'][0]['children'][0]['children'][2]['name'] == 'gitlabber-sample-submodule'


import enum


class CloneMethod(enum.IntEnum):
    SSH = 1
    HTTP = 2

    def __str__(self):
        return self.name.lower()

    def __repr__(self):
        return str(self)

    @staticmethod
    def argparse(s):
        try:
            return CloneMethod[s.upper()]
        except KeyError:
            return s



import os
import json
from gitlabber import __version__ as VERSION
import tests.gitlab_test_utils as gitlab_util
import tests.io_test_util as io_util
import pytest
import coverage
coverage.process_startup()

@pytest.mark.integration_test
def test_help():
    output = io_util.execute(["-h"])
    assert "usage:" in output
    assert "examples:" in output
    assert "positional arguments:" in output
    assert "optional arguments:" in output
    assert "Gitlabber - clones or pulls entire groups/projects tree from gitlab" in output

@pytest.mark.integration_test
def test_version():
    output = io_util.execute(["--version"])
    assert VERSION in output

@pytest.mark.integration_test
def test_file_input():
    os.environ['GITLAB_URL'] = 'http://gitlab.my.com/'
    output = io_util.execute(["-f", gitlab_util.YAML_TEST_INPUT_FILE, "-p", '-t', 'xxx'])
    with open(gitlab_util.TREE_TEST_OUTPUT_FILE, 'r') as treeFile:
        assert treeFile.read().strip() == output.strip()



from gitlabber.naming import FolderNaming
import pytest
import re

def test_naming_parse():
    assert FolderNaming.PATH == FolderNaming.argparse("PATH")

def test_naming_string():
    assert "name" == FolderNaming.__str__(FolderNaming.NAME)

def test_repr():
    retval = repr(FolderNaming.PATH)
    match = re.match("^<FolderNaming: ({.*})>$", retval)

def test_naming_invalid():
    assert "invalid_value" == FolderNaming.argparse("invalid_value")
        



import os
import sys
import subprocess
import sys
from contextlib import contextmanager
from io import StringIO


@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err

def execute(args, timeout=3):
    cmd = [sys.executable, '-m', 'gitlabber']
    cmd.extend(args)
    with subprocess.Popen(cmd, stdout=subprocess.PIPE, env=os.environ.copy()) as process:
        outs, err = process.communicate(timeout=timeout)    
        process.wait()
        return outs.decode('utf-8')



from gitlabber import cli
from gitlabber import __version__ as VERSION
import tests.io_test_util as output_util

from gitlabber.format import PrintFormat
from gitlabber.method import CloneMethod
from gitlabber.naming import FolderNaming
from gitlabber.archive import ArchivedResults
from unittest import mock
from anytree import Node
import pytest


def exit():
    import sys
    sys.exit()


def test_args_version():
    args_mock = mock.Mock()
    args_mock.return_value = Node(name="test", version=True)
    cli.parse_args = args_mock

    with output_util.captured_output() as (out, err):
        with pytest.raises(SystemExit):
            cli.main()
            assert VERSION == out.getvalue()


@mock.patch("gitlabber.cli.logging")
@mock.patch("gitlabber.cli.sys")
@mock.patch("gitlabber.cli.os")
@mock.patch("gitlabber.cli.log")
@mock.patch("gitlabber.cli.GitlabTree")
def test_args_logging(mock_tree, mock_log, mock_os, mock_sys, mock_logging):
    args_mock = mock.Mock()
    args_mock.return_value = Node(
        name="test", version=None, verbose=True, include="", exclude="", url="test_url", token="test_token", method=CloneMethod.SSH, naming=FolderNaming.PATH, archived=ArchivedResults.INCLUDE, file=None, concurrency=1, recursive=False, disble_progress=True, print=None, dest=".")
    cli.parse_args = args_mock

    mock_streamhandler = mock.Mock()
    mock_logging.StreamHandler = mock_streamhandler
    streamhandler_instance = mock_streamhandler.return_value
    mock_formatter = mock.Mock()
    streamhandler_instance.setFormatter = mock_formatter

    cli.main()

    mock_streamhandler.assert_called_once_with(mock_sys.stdout)
    mock_formatter.assert_called_once()


@mock.patch("gitlabber.cli.GitlabTree")
def test_args_include(mock_tree):
    inc_groups = "/inc**,/inc**"
    exc_groups = "/exc**,/exc**"
    args_mock = mock.Mock()
    args_mock.return_value = Node(
        name="test", version=None, debug=None, include=inc_groups, exclude=exc_groups, url="test_url", token="test_token", method=CloneMethod.SSH, naming=FolderNaming.NAME, archived=ArchivedResults.INCLUDE, file=None, concurrency=1, recursive=False, disble_progress=True, print=None, dest=".")
    cli.parse_args = args_mock

    split_mock = mock.Mock()
    cli.split = split_mock

    mock_tree.return_value.is_empty = mock.Mock(return_value=False)

    cli.main()
    split_mock.assert_has_calls([mock.call(inc_groups), mock.call(exc_groups)])


@mock.patch("gitlabber.cli.GitlabTree")
def test_args_include(mock_tree):
    args_mock = mock.Mock()
    args_mock.return_value = Node(
        name="test", version=None, verbose=None, include="", exclude="", url="test_url", token="test_token", method=CloneMethod.SSH, naming=FolderNaming.NAME, archived=ArchivedResults.INCLUDE, file=None, concurrency=1, recursive=False, disble_progress=True, print=True, dest=".", print_format=PrintFormat.YAML)
    cli.parse_args = args_mock

    print_tree_mock = mock.Mock()
    mock_tree.return_value.print_tree = print_tree_mock
    mock_tree.return_value.is_empty = mock.Mock(return_value=False)

    cli.main()

    print_tree_mock.assert_called_once_with(PrintFormat.YAML)


def test_validate_path():
    assert "/test" == cli.validate_path("/test/")
    assert "/test" == cli.validate_path("/test")
    assert "/" == cli.validate_path("//")
    assert "." == cli.validate_path("./")
    assert "." == cli.validate_path(".")

@mock.patch("gitlabber.cli.GitlabTree")
def test__missing_token(mock_tree):
    args_mock = mock.Mock()
    args_mock.return_value = Node(
        name="test", version=None, verbose=None, include="", exclude="", url="test_url", token=None, print=True, dest=".")
    cli.parse_args = args_mock

    with pytest.raises(SystemExit):
        cli.main()

@mock.patch("gitlabber.cli.GitlabTree")
def test_missing_url(mock_tree):
    args_mock = mock.Mock()
    args_mock.return_value = Node(
        name="test", version=None, verbose=None, include="", exclude="", url=None, token="some_token", print=True, dest=".")
    cli.parse_args = args_mock

    with pytest.raises(SystemExit):
        cli.main()
@mock.patch("gitlabber.cli.GitlabTree")
def test_empty_tree(mock_tree):
    args_mock = mock.Mock()
    args_mock.return_value = Node(
        name="test", version=None, verbose=None, include="", exclude="", url="test_url", token="test_token", method=CloneMethod.SSH, naming=FolderNaming.NAME, archived=ArchivedResults.INCLUDE, file=None, concurrency=1, recursive=False, disble_progress=True, print=True, dest=".")
    cli.parse_args = args_mock

    with pytest.raises(SystemExit):
        cli.main()


@mock.patch("gitlabber.cli.GitlabTree")
def test_missing_dest(mock_tree, capsys):
    args_mock = mock.Mock()
    args_mock.return_value = Node(
        name="test", version=None, verbose=None, include="", exclude="", url="test_url", token="test_token", method=CloneMethod.SSH, naming=FolderNaming.NAME, archived=ArchivedResults.INCLUDE, file=None, concurrency=1, recursive=False, disble_progress=True, print=False, dest=None)
    cli.parse_args = args_mock
    mock_tree.return_value.is_empty = mock.Mock(return_value=False)

    with pytest.raises(SystemExit):
        cli.main()
    out, err = capsys.readouterr()
    assert "Please specify a destination" in out



from gitlabber.format import PrintFormat
import pytest
import re

def test_format_parse():
    assert PrintFormat.JSON == PrintFormat.argparse("JSON")

def test_format_string():
    assert "json" == PrintFormat.__str__(PrintFormat.JSON)

def test_repr():
    retval = repr(PrintFormat.JSON)
    match = re.match("^<PrintFormat: ({.*})>$", retval)

def test_format_invalid():
    assert "invalid_value" == PrintFormat.argparse("invalid_value")
        



        

from tqdm import tqdm
import time
class ProgressBar:
    def __init__(self, description='', disabled=False):
        self.progress = None
        self.description = description
        self.disabled = disabled
        self.start = None

    def init_progress(self, total):
        if self.progress is None:
            self.start = time.time()
            self.progress = tqdm(total=total, unit="projects",
                                 bar_format="{desc}: {percentage:.1f}%|{bar:80}| {n_fmt}/{total_fmt}{postfix}", desc=self.description, leave=False, disable=self.disabled)

    def update_progress_length(self, added):
        if self.progress is not None:
            self.progress.total = self.progress.total + added
            self.progress.refresh()

    def show_progress(self, text, category='~'):
        if self.progress is not None:
            self.progress.update(1)
            postfix = {category : text}
            self.progress.set_postfix(postfix)

    def finish_progress(self):
        if self.progress is not None:
            self.progress.close()
            end = time.time() - self.start
            hours, rem = divmod(end-self.start, 3600)
            minutes, seconds = divmod(rem, 60)
            return "{:0>2}:{:0>2}:{:05.2f}".format(int(hours), int(minutes), seconds)



from gitlabber.method import CloneMethod
import pytest
import re

def test_method_parse():
    assert CloneMethod.SSH == CloneMethod.argparse("ssh")

def test_method_string():
    assert "http" == CloneMethod.__str__(CloneMethod.HTTP)

def test_repr():
    retval = repr(CloneMethod.SSH)
    match = re.match("^<CloneMethod: ({.*})>$", retval)


def test_method_invalid():
    assert "invalid_value" == CloneMethod.argparse("invalid_value")
        



import logging
import os
import sys
import subprocess
import git
from .progress import ProgressBar
import concurrent.futures

log = logging.getLogger(__name__)

progress = ProgressBar('* syncing projects')


class GitAction:
    def __init__(self, node, path, recursive=False):
        self.node = node
        self.path = path
        self.recursive = recursive

def sync_tree(root, dest, concurrency=1, disable_progress=False, recursive=False):
    if not disable_progress:
        progress.init_progress(len(root.leaves))
    actions = get_git_actions(root, dest, recursive)
    with concurrent.futures.ThreadPoolExecutor(max_workers=concurrency) as executor:
        executor.map(clone_or_pull_project, actions)
    elapsed = progress.finish_progress()
    log.debug("Syncing projects took [%s]", elapsed)


def get_git_actions(root, dest, recursive):
    actions = []
    for child in root.children:
        path = "%s%s" % (dest, child.root_path)
        if not os.path.exists(path):
            os.makedirs(path)
        if child.is_leaf:
            actions.append(GitAction(child, path, recursive))            
        if not child.is_leaf:
            actions.extend(get_git_actions(child, dest, recursive))
    return actions


def is_git_repo(path):
    try:
        _ = git.Repo(path).git_dir
        return True
    except git.InvalidGitRepositoryError:
        return False


def clone_or_pull_project(action):
    if is_git_repo(action.path):
        '''
        Update existing project
        '''
        log.debug("updating existing project %s", action.path)
        progress.show_progress(action.node.name, 'pull')
        try:
            repo = git.Repo(action.path)
            repo.remotes.origin.pull()
            if(action.recursive): 
                repo.submodule_update(recursive=True)
        except KeyboardInterrupt:
            log.fatal("User interrupted")
            sys.exit(0)
        except Exception as e:
            log.debug("Error pulling project %s", action.path, exc_info=True)
    else:
        '''
        Clone new project
        '''
        log.debug("cloning new project %s", action.path)
        progress.show_progress(action.node.name, 'clone')
        try:
            if(action.recursive):
                git.Repo.clone_from(action.node.url, action.path, multi_options=['--recursive'])
            else:
                git.Repo.clone_from(action.node.url, action.path)
        except KeyboardInterrupt:
            log.fatal("User interrupted")
            sys.exit(0)
        except Exception as e:
            log.debug("Error cloning project %s", action.path, exc_info=True)




"""Packaging settings."""


from codecs import open
import os
from os.path import abspath, dirname, join
from subprocess import call

from setuptools import Command, find_packages, setup

from gitlabber import __version__


this_dir = abspath(dirname(__file__))
with open(join(this_dir, 'README.rst'), encoding='utf-8') as file:
    long_description = file.read()


class RunTests(Command):
    """Run all tests."""
    description = 'run tests'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Run all tests!"""
        errno = call(['py.test'])
        raise SystemExit(errno)
setup(
    python_requires='>=3',
    name = 'gitlabber',
    packages = ['gitlabber'],
    version = __version__,
    description = 'A Gitlab clone/pull utility for backing up or cloning Gitlab groups',
    long_description = long_description,
    url = 'https://github.com/ezbz/gitlabber',
    download_url = 'https://github.com/ezbz/gitlabber/archive/master.zip',
    author = 'Erez Mazor',
    author_email = 'erezmazor@gmail.com',
    license = 'MIT',
    keywords = ['gitlab', 'python', 'cli'], 
    include_package_data = True,

    classifiers = [
        'Development Status :: 4 - Beta',   
        'Intended Audience :: Developers',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',     
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
    setup_requires = [
            'docopt', 
            'anytree', 
            'globre',
            'pyyaml',
            'tqdm',
            'GitPython', 
            'python-gitlab'
    ],
    install_requires = [
            'docopt', 
            'anytree', 
            'globre', 
            'pyyaml',
            'tqdm',
            'GitPython', 
            'python-gitlab'
    ],
    tests_require=  ['coverage', 'pytest', 'pytest-cov', 'pytest-integration'],
    entry_points = {
        'console_scripts': [
            'gitlabber=gitlabber.cli:main',
        ],
    },
    cmdclass = {'test': RunTests},
)



import pytest
import json
from unittest import mock
from gitlabber import gitlab_tree

URL = "http://gitlab.my.com/"
TOKEN = "MOCK_TOKEN"
GROUP_URL = "http://gitlab.my.com/group"
GROUP_NAME = "group"

SUBGROUP_URL = "http://gitlab.my.com/group/subgroup"
SUBGROUP_NAME = "subgroup"

PROJECT_URL = "http://gitlab.my.com/group/subgroup/project/project.git"
PROJECT_NAME = "project"

YAML_TEST_INPUT_FILE = "tests/test-input.yaml"
YAML_TEST_OUTPUT_FILE = "tests/test-output.yaml"
JSON_TEST_OUTPUT_FILE = "tests/test-output.json"
TREE_TEST_OUTPUT_FILE = "tests/test-output.tree"


class MockNode:
    def __init__(self, id, name, url, subgroups=mock.MagicMock(), projects=mock.MagicMock(), parent_id=None):
        self.id = id
        self.name = name
        self.path = name
        self.url = url
        self.web_url = url
        self.ssh_url_to_repo = url
        self.http_url_to_repo = url
        self.subgroups = subgroups
        self.projects = projects
        self.parent_id = parent_id


class Listable:
    def __init__(self, list_result, get_result=None, archive_result=None):
        self.list_result = list_result
        self.get_result = get_result
        self.archive_result = archive_result

    def list(self, as_list=False, archived=None):
        if archived is None:
            return [self.list_result, self.archive_result] if self.archive_result is not None else [self.list_result]
        elif archived is True:
            return [self.archive_result]
        else:
            return [self.list_result]

    def get(self, id):
        if self.get_result is not None:
            return self.get_result
        else:
            return self.list_result


def validate_root(root):
    assert root.is_leaf is False
    assert root.name == ""
    assert root.url == "http://gitlab.my.com/"
    assert len(root.children) == 1
    assert root.height == 3


def validate_group(group):
    assert group.name == GROUP_NAME
    assert group.url == GROUP_URL
    assert group.is_leaf is False
    assert len(group.children) == 1
    assert group.height == 2


def validate_subgroup(subgroup):
    assert subgroup.name == SUBGROUP_NAME
    assert subgroup.url == SUBGROUP_URL
    assert subgroup.is_leaf is False
    assert len(subgroup.children) == 1
    assert subgroup.height == 1


def validate_project(project):
    assert project.name == PROJECT_NAME
    assert project.url == PROJECT_URL
    assert project.is_leaf is True
    assert len(project.children) == 0


def validate_tree(root):
    validate_root(root)
    validate_group(root.children[0])
    validate_subgroup(root.children[0].children[0])
    validate_project(root.children[0].children[0].children[0])


def create_test_gitlab(monkeypatch, includes=None, excludes=None, in_file=None):
    gl = gitlab_tree.GitlabTree(
        URL, TOKEN, "ssh", "name", includes=includes, excludes=excludes, in_file=in_file)
    projects = Listable(MockNode(2, PROJECT_NAME, PROJECT_URL))
    subgroup_node = MockNode(2, SUBGROUP_NAME, SUBGROUP_URL, projects=projects)
    subgroups = Listable(subgroup_node)
    groups = Listable(MockNode(2, GROUP_NAME, GROUP_URL,
                               subgroups=subgroups), subgroup_node)
    monkeypatch.setattr(gl.gitlab, "groups", groups)
    return gl


def create_test_gitlab_with_toplevel_subgroups(monkeypatch):
    gl = gitlab_tree.GitlabTree(URL, TOKEN, "ssh", "path")
    groups = Listable([MockNode(2, GROUP_NAME, GROUP_URL),
                       MockNode(2, GROUP_NAME, GROUP_URL, parent_id=1)])
    monkeypatch.setattr(gl.gitlab, "groups", groups)
    return gl


def create_test_gitlab_with_archived(monkeypatch, includes=None, excludes=None, in_file=None, archived=None):
    gl = gitlab_tree.GitlabTree(
        URL, TOKEN, "ssh", "name", includes=includes, excludes=excludes, in_file=in_file, archived=archived)
    project_node = MockNode(1, PROJECT_NAME, PROJECT_URL)
    archived_project_node = MockNode(
        2, "_archived_" + PROJECT_NAME, "_archived_" + PROJECT_URL)
    projects = Listable(project_node, archive_result=archived_project_node)
    subgroup_node = MockNode(2, SUBGROUP_NAME, SUBGROUP_URL, projects=projects)
    archived_subgroup_node = MockNode(
        2, "_archived_" + SUBGROUP_NAME, "_archived_" + SUBGROUP_URL, projects=projects)
    subgroups = Listable(subgroup_node, archive_result=archived_subgroup_node)
    archived_subgroups = Listable(archived_subgroup_node, archive_result=archived_subgroup_node)
    group_node = MockNode(2, GROUP_NAME, GROUP_URL, subgroups=archived_subgroups)
    archived_group_node = MockNode(2, "_archived_" + GROUP_NAME, "_archived_" + GROUP_URL, subgroups=archived_subgroups)
    groups = Listable(group_node, get_result=subgroup_node, archive_result=archived_group_node)
    monkeypatch.setattr(gl.gitlab, "groups", groups)
    # gl.print_tree()
    return gl





from gitlabber import git
from gitlabber.git import GitAction
from unittest import mock
from anytree import Node
import pytest

DEST="./test_dest"
GROUP_PATH = "/group"
SUBGROUP_PATH = "/group/subgroup"
PROJECT_PATH = "/group/subgroup/project"

def create_tree():
    root = Node(name="root")
    group = Node(name="group", root_path=GROUP_PATH, parent=root)
    subgroup = Node(name="subgroup", root_path=SUBGROUP_PATH, parent=group)
    Node(name="project1", root_path=PROJECT_PATH, parent=subgroup)
    return root


@mock.patch('gitlabber.git.os')
@mock.patch('gitlabber.git.git')
@mock.patch('gitlabber.git.clone_or_pull_project')
@mock.patch('gitlabber.git.progress')
def test_create_new_user_dir(mock_progress, mock_clone_or_pull_project, mock_git, mock_os):
    # git.git = mock.MagicMock()
    
    mock_os.path.exists.return_value = False

    root = create_tree()
    git.sync_tree(root,DEST)
    
    assert 3 == mock_os.path.exists.call_count
    mock_os.path.exists.assert_has_calls(
        [mock.call(DEST+GROUP_PATH), mock.call(DEST+SUBGROUP_PATH), mock.call(DEST+PROJECT_PATH)])

    assert 3 == mock_os.makedirs.call_count
    mock_os.makedirs.assert_has_calls(
        [mock.call(DEST+GROUP_PATH), mock.call(DEST+SUBGROUP_PATH), mock.call(DEST+PROJECT_PATH)])

    assert 1 == git.clone_or_pull_project.call_count


@mock.patch('gitlabber.git.git')
def test_is_git_repo_true(mock_git):
    mock_repo = mock.Mock()
    mock_git.Repo = mock_repo
    git.is_git_repo("dummy_dir")
    assert 1 == mock_git.Repo.call_count
    mock_git.Repo.assert_called_once_with("dummy_dir")


def test_is_git_repo_throws():
    with pytest.raises(git.git.exc.NoSuchPathError):
        git.is_git_repo("dummy_dir")

@mock.patch('gitlabber.git.git')
def test_pull_repo(mock_git):
    mock_repo = mock.Mock()
    mock_git.Repo = mock_repo
    repo_instance = mock_git.Repo.return_value
    git.is_git_repo = mock.MagicMock(return_value=True)

    git.clone_or_pull_project(GitAction(Node(name="test"), "dummy_dir"))
    mock_git.Repo.assert_called_once_with("dummy_dir")
    repo_instance.remotes.origin.pull.assert_called_once()


@mock.patch('gitlabber.git.git')
def test_clone_repo(mock_git):
    mock_repo = mock.Mock()
    mock_git.Repo = mock_repo
    git.is_git_repo = mock.MagicMock(return_value=False)

    git.clone_or_pull_project(
        GitAction(Node(name="dummy_url", url="dummy_url"), "dummy_dir"))

    mock_git.Repo.clone_from.assert_called_once_with("dummy_url", "dummy_dir")

@mock.patch('gitlabber.git.git')
def test_clone_repo_recursive(mock_git):
    mock_repo = mock.Mock()
    mock_git.Repo = mock_repo
    git.is_git_repo = mock.MagicMock(return_value=False)

    git.clone_or_pull_project(
        GitAction(Node(name="dummy_url", url="dummy_url"), "dummy_dir", recursive=True))

    mock_git.Repo.clone_from.assert_called_once_with("dummy_url", "dummy_dir", multi_options=['--recursive'])

@mock.patch('gitlabber.git.git')
def test_pull_repo_recursive(mock_git):
    mock_repo = mock.Mock()
    mock_git.Repo = mock_repo
    repo_instance = mock_git.Repo.return_value
    git.is_git_repo = mock.MagicMock(return_value=True)

    git.clone_or_pull_project(GitAction(Node(name="test"), "dummy_dir", recursive=True))
    mock_git.Repo.assert_called_once_with("dummy_dir")
    repo_instance.remotes.origin.pull.assert_called_once()
    repo_instance.submodule_update.assert_called_once_with(recursive=True)

@mock.patch('gitlabber.git.git')
def test_pull_repo_interrupt(mock_git):
    mock_repo = mock.Mock()
    mock_git.Repo = mock_repo
    git.is_git_repo = mock.MagicMock(return_value=True)

    repo_instance = mock_git.Repo.return_value
    repo_instance.remotes.origin.pull.side_effect=KeyboardInterrupt('pull test keyboard interrupt')

    with pytest.raises(SystemExit):
        git.clone_or_pull_project(GitAction(
            Node(name="dummy_url", url="dummy_url"), "dummy_dir"))

    mock_git.Repo.assert_called_once_with("dummy_dir")
    repo_instance.remotes.origin.pull.assert_called_once()

@mock.patch('gitlabber.git.git')
def test_clone_repo_interrupt(mock_git):
    mock_repo = mock.Mock()
    mock_git.Repo = mock_repo
    git.is_git_repo = mock.MagicMock(return_value=False)
    mock_git.Repo.clone_from.side_effect=KeyboardInterrupt('clone test keyboard interrupt')

    with pytest.raises(SystemExit):
        git.clone_or_pull_project(GitAction(
            Node(name="dummy_url", url="dummy_url"), "dummy_dir"))

    mock_git.Repo.clone_from.assert_called_once_with("dummy_url", "dummy_dir")
